#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>

#include <rlcif.h>

#include "receiver.h"
#include "agent.h"
#include "rmdpif.h"
#include "fec.h"
#include "rmdp.h"
#include "init.h"
#include "utils.h"

#define getinstance(i) ((rmdp_inst_t *)_rlc_getinstance(i, rmdp_inst_mngr))
static int receiver_init_done = 0;

/* this is an hack ... I'm using rlc net.o */
int openssock(unsigned long addr, short port, int ttl, int noblock);
 
static int decode(rmdp_inst_t *inst);
static int save(rmdp_inst_t *inst);
static int send_request(void *ides);
static int ask_more(rmdp_inst_t *inst, int *rtime);
static int setup_feed_sock (rmdp_inst_t *inst);
static void unset_feed_sock (rmdp_inst_t *inst);

#define RMDP_RETR_TIM	9090
#define KFACT		0.2	/* moving average factor */
#define SAFEF		1.2	/* safety factor for extension */

/*++++
 Sets the function to be called when the reception terminates because of
 a timeout. This function is called either if the receiver times out
 waiting for the next packet, or if it is unable to prolong a too short
 transmission. If this timeout handler is not set, the usual
 asynchronous error handling mechanism is used instead.

Params:
    - pi: protocol instance identifier.
    - err_f: the error callback function. It is of type
      typedef void (*rmdp_rx_err_f) (int pi);
 
Returns:
    - 0 on success;
    - a negative value on failure:
*/
int
rmdp_rx_set_err_callback (int pi, rmdp_rx_err_f err_f)
{
  rmdp_inst_t *inst;

    if ((inst = getinstance(pi)) == NULL) {
	return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }
    inst->err_f = err_f;

    return (0);
}

/*++++
 Enables multicast feedback. Default is unicast!

 Params:
	- pi: protocol instance identifier.
	- multifeed: if != enable multicast, otherwise unicast.

 Returns:
	- 0 on success, a negative value on failure.
*/
int rmdp_rx_enable_multifeed (int pi, int multifeed)
{
  rmdp_inst_t *inst;

    if ((inst = getinstance(pi)) == NULL) {
	return (rlc_return_error_(pi, RMDP_NOSUCH_INST, NULL));
    }
    if (multifeed)
	inst->mcastfeed = 1;
    else
	inst->mcastfeed = 0;

    unset_feed_sock (inst);

    return (0);
}

int _rmdp_receiver_init()
{
    if (!receiver_init_done) {
	receiver_init_done = 1;
    }

    return (0);
}

void
_rmdp_send_stats (int pi, void *buff, int size)
{
  char tbu[RMDP_STATS_PAYSIZE+sizeof(rmdp_feed_pkt_t)];
  rmdp_inst_t *inst;
  int thissize;

    DEB(fprintf(stderr, "_rmdp_send_stats pi %d with size %d\n", pi, size);)
    if ((inst = getinstance(pi)) == NULL) {
	/* rlc_have_error_(-1, RMDP_NOSUCH_INST, NULL); */
	return;
    }

    /* XXX */
    thissize = min (size, RMDP_STATS_PAYSIZE);
    bcopy (buff, tbu+sizeof(rmdp_feed_pkt_t), thissize);

    if (inst->mfeeso == -1 && inst->ufeeso == -1 && setup_feed_sock (inst)
	< 0) {
	/* rlc_have_error_ (inst->rlc_pi, RMDP_NO_SOCK, NULL); */
	return;
    }

    RMDP_FEE_RLC_FILL(*((rmdp_feed_pkt_t *)tbu), inst->sessid);
    if (send(inst->mfeeso, (char *)&tbu, thissize, 0) == -1) {
	/* rlc_have_error_ (inst->rlc_pi, RMDP_ER_SND, NULL); */
	return;
    }
}

/* reveived paket handling */
void
_rmdp_rx_packet(int ides, int lay, char flags, char *pkt,
	int size)
{
  int i, ret;
  int pkt_n, blk_n;
  rmdp_inst_t *inst;
  int rtime, newleft;
    NDEB(fprintf(stderr,"*** got l %d, siz %d\n", lay, size);)
    if (size <= 0) {
NDEB(fprintf(stderr, "--- rx_packet: (lay %d) skipping void (%d) pkt\n",
		lay, size);)
	return;
    }

    if ((inst = getinstance(ides)) == NULL) {
	rlc_have_error_(-1, RMDP_NOSUCH_INST,
	    "* rx_packet called for unknown instance");
	return;
    }
    
    if (inst->rlc_pi < 0) {
	rlc_have_error_(-1, RMDP_NO_RLC_ID, NULL);
	return;
    }

    /* see what we have */
    pkt_n = ntohs(RMDP_H_P(pkt)->the_k);
    blk_n = ntohs(RMDP_H_P(pkt)->the_B);

    NDEB(fprintf(stderr,"--- got B %d, k %d\n", pkt_n, blk_n);)
    /* must allocate data buffers */
    if (inst->B == 0 ) {
	inst->k = ntohs(RMDP_H_P(pkt)->k);
	inst->B = ntohs(RMDP_H_P(pkt)->B);
	inst->flength = ntohl(RMDP_H_P(pkt)->flength);
	inst->psize = size - RMDP_H_S;
	inst->npkts = (inst->flength+inst->psize-1)/inst->psize;
	NDEB(fprintf(stderr,"--- first packet, must allocate things"
		" (len %lu B %d)\n", inst->flength, inst->B);)
	NDEB(fprintf(stderr,"--- got B %d, k %d\n", inst->B, inst->k));
	inst->missingblocks = inst->B;
	inst->duppackets = 0;
	if ( (inst->B > 4000) || (inst->flength > 10000000) ) {
	    rlc_have_error_(inst->rlc_pi, RMDP_DATA_SIZE, NULL);
	    /* _rmdp_deleteinstance(inst); */
	    free (pkt);	/* is it OK? XXX */
	    return;
	}
	if (_rmdp_allocation(inst) < 0) {
	    rlc_have_error_(inst->rlc_pi, RMDP_ALLOC_ERROR, NULL);
	    _rmdp_deallocation(inst);
	    free (pkt);	/* is it OK? XXX */
	    return;
	}
	if (inst->stafd > 0 && (ret = rlc_rx_set_stats_deliver_f (
	    inst->rlc_pi, ides, _rmdp_send_stats, RMDP_STATS_PAYSIZE)) < 0)
	    rlc_have_error_(inst->rlc_pi, ret, NULL);
	
#ifdef HAVE_UI
	if (inst->have_gui) {
	    rlc_gui_tcl_exec(inst->rlc_pi, NULL,
		"pkts_init %d %d %d %d",
		    inst->B, inst->k, inst->flength, inst->psize);
	    NDEB(fprintf(stderr,"--- UI update: "
		"pkts_init %d %d\n", inst->B, inst->k);)
	}
#endif
    }

    /* if have sessid check if it belongs to us !! */
    if (inst->sessid && inst->sessid != ntohl(RMDP_H_P(pkt)->sessid)) {
	free (pkt);	/* is it OK? XXX */
	return;
    } else {
	inst->sessid = ntohl(RMDP_H_P(pkt)->sessid);
    }

    inst->totgot ++;

    /* update a stat .. */
    newleft = ntohl(RMDP_H_P(pkt)->left);
    NDEB(fprintf(stderr, "now left is %d\n", newleft);)
    if (inst->left > newleft) {
	int f;
	f = (inst->totgot-inst->oldtotgot);
	inst->oldtotgot = inst->totgot;
	inst->speedf += KFACT * (f - inst->speedf);
	NDEB(fprintf(stderr, "now f is %f\n", inst->speedf);)
    }

    if (inst->countdata[blk_n] >= inst->k) {
	inst->duppackets++;
	free (pkt);	/* is it OK? XXX */
    } else {
	int valid = 1;
	for (i=0; i<inst->countdata[blk_n]; i++)
	    if (pkt_n == inst->whichdata[blk_n][i]) {
		valid = 0;
	NDEB(fprintf(stderr, "not valid.. blk %d pkt %d\n",
		    blk_n, pkt_n))
#ifdef HAVE_UI
		if (inst->have_gui) {
		    rlc_gui_tcl_exec(inst->rlc_pi, NULL,
			"pkts_have_dup %d 1", blk_n);
		    NDEB(fprintf(stderr,"--- UI update: pkts_have_dup %d 1\n",
			blk_n);)
		}
#endif
		break;
	    }
	if (valid) {
	    inst->netgot ++;
	    inst->whichdata[blk_n][inst->countdata[blk_n]] = pkt_n;
	    inst->maxn = max (inst->maxn, pkt_n);
	    inst->alldata[blk_n][inst->countdata[blk_n]] = pkt;
	    inst->allpkts[blk_n][inst->countdata[blk_n]] = RMDP_DATA(pkt);
	    inst->countdata[blk_n]++; 
	    NDEB(fprintf(stderr, "now blk %4d pkt %4d (%4d pkts); "
		    "miss %4d pkts %4d blks\n",
		    blk_n, pkt_n, inst->countdata[blk_n],
		    inst->k-inst->countdata[blk_n],inst->missingblocks))
	    if (inst->countdata[blk_n] == inst->k) {
	        inst->missingblocks--;
	    }
#ifdef HAVE_UI
	    if (inst->have_gui) {
		int part = inst->netgot * 50 / (inst->B * inst->k);
		if (part > inst->recpercent) {
		    inst->recpercent = part;
		    rlc_gui_tcl_exec(inst->rlc_pi, NULL, "prog_percent %f",
			part/50.0);
		}
	    }
#endif
	} else {
	    inst->duppackets++;
	    free (pkt);	/* is it OK? XXX */
	}
    }
#ifdef HAVE_UI
    if (inst->have_gui) {
	rlc_gui_tcl_exec(inst->rlc_pi, NULL, "pkts_have_more %d 1", blk_n);
	NDEB(fprintf(stderr,"--- UI update: pkts_have_more %d 1\n", blk_n);)
    }
#endif

    /* end reception ? */
    if (!inst->missingblocks) {
	int res;
	inst->reqpending = 0;
	res = decode(inst);
	NDEB(fprintf(stderr, "rmdp: h_p %d (%x) decoded with %d\n",
	    ides, (int)inst, res);)
	if (inst->end_f)
	    inst->end_f(ides, res);
	NDEB(fprintf(stderr, "rmdp: h_p after end_f for %d (%x)\n", ides,
	    (int)inst);)
	save(inst);
	/* _rmdp_deleteinstance(inst); */
	NDEB(fprintf(stderr, "rmdp: h_p delated %d (%x)\n", ides, (int)inst);)
	return;
    }

    /* handle extension of transmission */
     
    /* sender has extended the trnasmissin,
     * we try to `reset' the timeout process */
    if (inst->reqpending && newleft > inst->left)
	inst->reqsent = 0;

    inst->left = newleft;
    if ((!inst->reqpending) && (inst->reqpending=ask_more(inst, &rtime))) {
	inst->reqpending = 1;
	inst->reqsent = 0;
	NDEB(fprintf(stderr, "scheduling req at %d\n", rtime);)
	if (rlc_settimer(RMDP_RETR_TIM, rtime, (CALLBACK)send_request,
	    (void *)ides, 0) < 0) {
	    rlc_have_error_(inst->rlc_pi, RMDP_ER_RETR_TI, NULL);
	}
    } else if (inst->reqpending && !ask_more(inst, &rtime)) {
	/* the sender has possibly extended the transmission */
	inst->reqpending = 0;
	inst->reqsent = 0;
    }

}

/* If we should ask for an extension return the amount,
 * otherwise return 0.
 */
static int
ask_more(rmdp_inst_t *inst, int *rtime)
{
  int more;
  rx_info_t info;
  double adv;
  int len = sizeof(rx_info_t);

    /*
    ** B*k is the number of `net' packets we want
    ** netgot is the amount we got
    ** totgot is the total number we got
    ** speedf is the # pkts got for each pkt sent on lay 0
    */
    more = ((
		(inst->B * inst->k * SAFEF)	/*tot safly need*/	-
		inst->netgot			/*got this so far*/
	   )					/*=net needed*/		*
	   inst->netgot/(double)inst->totgot	/*waste factor*/	
	   					/*=tot needed*/		-
	   inst->left				
	   )					/*=will be needed*/	/
	   (double)inst->speedf + 1;		/*=pkt on 0 will be needed*/

    /* prevent asking for a small increment at
     * the begginning of the reception */
    rlc_rx_info(inst->rlc_pi, (void *)&info, &len);
    adv = 1. - (double)inst->left/(inst->B * inst->k);
    if (adv < .75 && (info.t0 * inst->left) > 2e6 && 
	more < (inst->B * inst->k - inst->netgot))
	more = 0;

    if (more > 0) {
	/* XXX need some error checking ... */
	*rtime = info.t0 * inst->left * drand48() / 2;
	if (*rtime <= 0)
	    *rtime =  RMDP_LAST_RETR_TIM;
	NDEB(fprintf(stderr, "ask_more ret %d (%d) \n", more, *rtime);)
	return(more);
    }
    
    return (0);
}

/* send a continuation request */
static int
send_request(void *ides)
{
  static rmdp_cont_pkt_t cp;
  int rtime;
  unsigned long howmany;
  rmdp_inst_t *inst;

   NDEB(fprintf(stderr, "*** send_request called \n");)

    /* always uses mfeeso (both for uni and multi) */

    if ((inst = getinstance((int)ides)) == NULL) {
	return 0;
    }

    if (inst->mfeeso == -1 && inst->ufeeso == -1 && setup_feed_sock (inst) < 0)
	rlc_have_error_ (inst->rlc_pi, RMDP_NO_SOCK, NULL);

    if (!inst->reqpending)
	return 0;

    if (inst->reqsent >= RMDP_MAX_RETRIES) {
	if (inst->err_f) {
	    inst->err_f( (int)ides );
	} else {
	    rlc_have_error_(inst->rlc_pi, RMDP_ER_TIMEOUT, NULL);
	}
	/* _rmdp_deleteinstance(inst); */
	return 0;
    }

    howmany = ask_more(inst, &rtime); 
    RMDP_CP_FILL(cp, inst->sessid, howmany, inst->left);
    if (send(inst->mfeeso, (char *)&cp, RMDP_CP_S, 0) == -1) {
	rlc_have_error_ (inst->rlc_pi, RMDP_ER_SND, NULL);
    }
    inst->reqsent++;
    
    if (inst->reqsent < RMDP_MAX_RETRIES) {
	if (rlc_settimer(RMDP_RETR_TIM, rtime, (CALLBACK)send_request,
	    (void *)ides, 0) < 0) {
	    rlc_have_error_(inst->rlc_pi, RMDP_ER_RETR_TI, NULL);
	}
    } else {
	if (rlc_settimer(RMDP_RETR_TIM, RMDP_LAST_RETR_TIM,
	    (CALLBACK)send_request, (void *)ides, 0) < 0) {
	    rlc_have_error_(inst->rlc_pi, RMDP_ER_RETR_TI, NULL);
	}
    }

    return 0;
}

static int
save(rmdp_inst_t *inst)
{
    int i, j, len = inst->psize;
    for (i=0; i<inst->B; i++) {
	for (j=0; j<inst->k && len > 0; j++) {
	    len=inst->psize;
	    if (inst->flength < len) len=inst->flength;
	    inst->flength -= len;
	    write(inst->fd, inst->allpkts[i][j], len);
	}
    }
    return 0;
}

static int
decode(rmdp_inst_t *inst)
{
    int i, thisn;
    /* first build allocate a `decoder' */
    thisn = max (256, inst->maxn+1);
    if ((inst->feccode = get_fec(inst->k, thisn)) == NULL) {
	rlc_have_error_(inst->rlc_pi, RMDP_COD_INST_ERR, NULL);
	/* _rmdp_deleteinstance(inst); */
	return -1;
    }

    NDEB (fprintf(stderr, "have decoder with %d %d\n", inst->k, thisn);)
    for (i=0; i<inst->B; i++) {
	NDEB (for(j=0; j<inst->k; j++)
	    fprintf(stderr, "%d ", inst->whichdata[i][j]);
	    fprintf(stderr, " indexes\n");
	)
	if (fec_decode(inst->feccode, inst->allpkts[i], inst->whichdata[i],
	    inst->psize) != 0) {
	    rlc_have_error_(inst->rlc_pi, RMDP_DECODER_ERR, NULL);
	    /* _rmdp_deleteinstance(inst); */
	    return -1;
	}
    }
    return 0;
}

static int
setup_feed_sock (rmdp_inst_t *inst)
{
  unsigned long srcaddr;

    /* always uses mfeeso (both for uni and multi) */
    if (inst->mcastfeed) {
	if ((inst->mfeeso =
	    openssock(inst->addr, inst->port,  inst->ttl, 0)) < 0 ) {
	    return (rlc_return_error_ (inst->rlc_pi, RMDP_NO_SOCK, NULL));
	}
    } else if ((srcaddr = rlc_rx_get_src(inst->rlc_pi)) != -1) {
	if ((inst->mfeeso =
	    openssock(srcaddr, inst->port,  inst->ttl, 0)) < 0 ) {
	    return (rlc_return_error_ (inst->rlc_pi, RMDP_NO_SOCK, NULL));
	}
    } else {
	return (rlc_return_error_ (inst->rlc_pi, RMDP_NO_SOCK,
	    "Cannot send feedback: don't know sender address"));
    }

    return 0;
}

static void
unset_feed_sock (rmdp_inst_t *inst)
{
    if (inst->mfeeso != -1) {
	close(inst->mfeeso);
	inst->mfeeso = -1;
    }
    if (inst->ufeeso != -1) {
	close(inst->ufeeso);
	inst->ufeeso = -1;
    }
}
