#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>
#include <fcntl.h>

#include <rlcif.h>

#include "rmdpif.h"
#include "fec.h"
#include "rmdp.h"
#include "layers.h"
#include "sender.h"
#include "receiver.h"
#include "agent.h"
#include "init.h"
#include "utils.h"

#define getinstance(i) ((rmdp_inst_t *)_rlc_getinstance(i, rmdp_inst_mngr))

#ifdef HAVE_UI
extern char RMDP_TCL_SCRIPT[];
extern char RMDP_RX_TCL_SCRIPT[];
extern char RMDP_TX_TCL_SCRIPT[];
#endif

/*++++
 
Creates a new protocol instance assigning a local protocol instance
identifier.

Params:
    - isreceiver: 1 means that this is a receiver, 0 a sender.
 
Returns:
    - the protocol instance id, on success.
    - a negative value on failure:
 
*/
 
int
rmdp_new (int isreceiver)
{
  int ret;
  rmdp_inst_t *se;

    /*
    ** global library init
    */
    if (rmdp_init_() < 0)
	return (rlc_return_error_(-1, RMDP_INITERR, NULL));

    /*
    ** sender init
    */
    if (_rmdp_sender_init() < 0)
	return (rlc_return_error_(-1, RMDP_INITERR, NULL));

    /*
    ** receiver init
    */
    if (_rmdp_receiver_init() < 0)
	return (rlc_return_error_(-1, RMDP_INITERR, NULL));

    /**
     ** allocate session descriptor
     **/
    if (NULL == (se = calloc(1, sizeof(rmdp_inst_t)))) {
	return (rlc_return_error_ (-1, RMDP_ALLOC_ERROR,
	    "* rmdp: Cannot allocate session descriptor"));
    }

    if (isreceiver) {
	se->sessid = 0;		/* 0 means not assigned, if start receiving
				 * like that, it will pick up the sessid in
				 * the first valid packet and stick to that */
    } else {
	se->sessid = rmdp_compute_sessid_();
    }
    se->alldata = NULL;
    se->B = se->k = se->npkts = se->flength = 0;
    se->psize = RMDP_PKT_SIZE;	/* in the receiver will be set to the right
				 * value at the first received packet */
    se->feccode = NULL;
    se->rlc_pi = -1;
    se->have_gui = 0;
    se->addr = 0;
    se->port = 0;
    se->ufeeso = -1;
    se->mfeeso = -1;
    se->fd = -1;
    if (isreceiver)
	se->isreceiver = 1;
    else
	se->isreceiver = 0;

    /* sender specific ... */
    se->nlay = RMDP_NLAY;
    se->idxs = NULL; se->allpkts = NULL; se->sparepkt = NULL;
    se->pkts_left = se->init_left = 0;
    if ((se->idxs = (unsigned short *)calloc(se->nlay,
	sizeof(unsigned short))) == NULL) {
	_rmdp_deleteinstance(se);
	return (rlc_return_error_(-1, RMDP_ALLOC_ERROR,
	    "* rmdp: Cannot file data structure"));
    }
    se->stafd = -1;


    /* receiver specific ... */
    se->ttl = 0;
    se->whichdata = NULL; se->countdata = NULL;
    se->maxn = se->missingblocks = se-> duppackets = 0;
    se->oldtotgot = se->totgot = se->netgot = 0;
    se->speedf = 1.;
    se->end_f = NULL; se->err_f = NULL;
    se->reqpending = se->reqsent = 0;
    se->recpercent = 0;
    se->left = 0;
    se->mcastfeed = 0;		/* default unicast ! */

    if ((ret = _rlc_insert_instance ((instance_des_t *)se, rmdp_inst_mngr))
	< 0) {
	_rmdp_deleteinstance(se);
	return (rlc_return_error_(-1, RMDP_TOOMANY_INST, "rmdp Error: too many"
	    " instances"));
    }

    return (ret);
}

/*++++
 
Provides and input/output descriptor associated with the file
being transmitted/received (sender/receiver). In the sender the input
descriptor must be able to provide the file length, using fstat().

Params:
    - pi: protocol instance identifier.
    - fd: input/output file descriptor.
 
Returns:
    - 0 on success;
    - a negative value on failure:
 
*/
 
int
rmdp_set_io (int pi, int fd)
{
  int ret;
  rmdp_inst_t *se;

    if ((se = getinstance(pi)) == NULL) {
	return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }

    /**
     ** chek output file descriptor
     **/
    if (fd < 0) {
	return (rlc_return_error_ (se->rlc_pi, RMDP_OUT_FILE_ERR, NULL));
    }

    if (!se->isreceiver && (ret = _rmdp_tx_read(pi, fd)) < 0) {
	return (rlc_return_error_(se->rlc_pi, ret, NULL));
    }

    se->fd = fd;

    return (0);
}

/*++++
 
Provides multicast address/port for the feedback channel.

Params:
    - pi: protocol instance identifier.
    - adds: data structures defined in "rmdpif.h", containing
      address (host order), port and ttl (used in the receiver only).
 
Returns:
    - 0 on success;
    - a negative value on failure:
 
*/
 
int
rmdp_set_addr (int pi, rmdp_addr_t *adds)
{
  rmdp_inst_t *se;

    if ((se = getinstance(pi)) == NULL) {
	return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }

    if (!IN_CLASSD(adds->addr))
	return rlc_return_error_(se->rlc_pi, RMDP_NOMCAST_ADDR, NULL);

    se->addr = htonl(adds->addr);
    se->port = htons(adds->port);
    se->ttl = adds->ttl;

#if 0
    if ( se->isreceiver &&
	(se->mfeeso = openssock(se->addr, se->port,  adds->ttl, 0)) < 0 ) {
	return (rlc_return_error_ (se->rlc_pi, RMDP_NO_SOCK, NULL));
    }
#endif

    return (0);
}

/*++++
 Set the flow instance id.

 In the sender, the instance id must be unique within the sender host.
 If not set using this function, it will be generated internally.

 In the receiver, only incoming packet that belong to this instance
 will be considered. It should be called before starting the reception.
 If the instance id is not set, the one found in the first valid packet
 will be used.

Params:
    - pi: protocol instance identifier.
    - id: the flow instance id.
 
Returns:
    - 0 on success;
    - a negative value on failure:
 
*/
int
rmdp_set_id (int pi, unsigned long id)
{
  rmdp_inst_t *inst;

    if ((inst = getinstance(pi)) == NULL) {
	return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }
    inst->sessid = id;

    return (0);
}

/*++++
 Get the flow instance id.

Params:
    - pi: protocol instance identifier.
    - id: ponter to a memory area that will be set to
      the value of the flow instance id.
 
Returns:
    - 0 on success;
    - a negative value on failure:
 
*/
int
rmdp_get_id (int pi, unsigned long *id)
{
  rmdp_inst_t *inst;

    if ((inst = getinstance(pi)) == NULL) {
	return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }
    *id = inst->sessid;

    return (0);
}

/*++++
 Retrieves a reference to the function that gets called to provide
 the outgoing/consume the incoming data packets.

 The function returned will be of type
    typedef void (* rmdp_rx_packet_t)
        (int pi, int lay, char flags, char *p, int size);
 in the receiver, and
    typedef int (* rmdp_tx_packet_t)
        (int pi, int lay, char flags, char **p);
 in the sender.

Params:
    - pi: protocol instance identifier.
    - f: ponter to a memory area that will be set to
      the ponter of the function.
 
Returns:
    - 0 on success;
    - a negative value on failure:
*/
int
rmdp_get_f (int pi, void **f)
{
  rmdp_inst_t *inst;

    if ((inst = getinstance(pi)) == NULL) {
        return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }

    if (inst->isreceiver)
	*f = (void *)_rmdp_rx_packet;
    else
	*f = (void *)_rmdp_tx_packet;

    return (0);
}


/*++++
 Deletes a protocol instance and claims back the storage used.

Params:
    - pi: protocol instance identifier.

Returns:
    - 0 on success;
    - a negative value on failure:

*/
int
rmdp_delete (int pi)
{
    _rmdp_deleteinstance(getinstance(pi));

    return (0);
}

/*++++
 Retrieves the rmdp packet size (rmdp payload size + rmdp header).

Params:
    - pi: protocol instance identifier.
 
Returns:
    - the packet size on success;
    - a negative value on failure:

*/
int
rmdp_get_psize (int pi)
{
  rmdp_inst_t *inst;

    if ((inst = getinstance(pi)) == NULL) {
	return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }

    return ( inst->psize + RMDP_H_S );
}

/*++++
 Registers the associated rlc instance ID. This function is ininfluent
 in the sender.

Params:
    - pi: protocol instance identifier.
 
Returns:
    - 0 on success;
    - a negative value on failure:
 
*/
int
rmdp_set_rlc (int pi, int rlc_pi)
{
  rmdp_inst_t *inst;
 
    if ((inst = getinstance(pi)) == NULL) {
	return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }
 
    inst->rlc_pi = rlc_pi;
    return (0);
}

/*++++
 Gets the associated rlc instance ID, previously set with
 rmdp_set_rlc.

Params:
    - pi: protocol instance identifier.
    - id: ponter to a memory area that will be set to
      the value of the rlc instance id.
 
Returns:
    - 0 on success;
    - a negative value on failure:
*/
int
rmdp_get_rlc (int pi, unsigned long *id)
{
  rmdp_inst_t *inst;
 
    if ((inst = getinstance(pi)) == NULL) {
        return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }

    *id = inst->rlc_pi;
 
    return 0;
}

/*++++
 Enables the UI capability. Use only after
 rmdp_set_rlc is called.
 It has no effect unless both rmdp and
 rlc have been compiled with the proper flag (HAVE_UI).

Params:
    - pi: protocol instance identifier.
Returns:
    - 0 on success;
    - a negative value on failure:
*/
int
rmdp_enable_gui (int pi)
{
  rmdp_inst_t *inst;

    if ((inst = getinstance(pi)) == NULL) {
	return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }

#ifdef HAVE_UI
    if ( rlc_gui_loadtcl(inst->rlc_pi, RMDP_TCL_SCRIPT) < 0 ||
	(inst->isreceiver &&
	 rlc_gui_loadtcl(inst->rlc_pi, RMDP_RX_TCL_SCRIPT) < 0) ||
	(!inst->isreceiver &&
	 rlc_gui_loadtcl(inst->rlc_pi, RMDP_TX_TCL_SCRIPT) < 0) ) {
	return (rlc_return_error_(inst->rlc_pi, RMDP_NO_UI, NULL));
    } else
	inst->have_gui = 1;

    DEB(fprintf(stderr, "Yes, we have GUI\n");)
    return (0);
#else
    return (rlc_return_error_(inst->rlc_pi, RMDP_NO_UI, NULL));
#endif
}

/*++++
Sets the application name in the UI. Use only after
rmdp_set_rlc is called.

Params:
    - pi: protocol instance identifier.
    - name: the application name.
 
Returns:
    - 0 on success;
    - a negative value on failure:
*/
int
rmdp_set_name (int pi, char *name)
{
  rmdp_inst_t *inst;
#ifdef HAVE_UI
  char *tmp;
#endif

    if ((inst = getinstance(pi)) == NULL) {
	return (rlc_return_error_ (-1, RMDP_NOSUCH_INST, NULL));
    }

#ifdef HAVE_UI
    if (!inst->have_gui)
	return (rlc_return_error_ (inst->rlc_pi, RMDP_NO_UI, NULL));

    if (inst->rlc_pi == -1)
	return (rlc_return_error_ (inst->rlc_pi, RMDP_NO_RLC_ID, NULL));

    if ((tmp = malloc(strlen(name) + strlen("wm title . \"rmdp: ") +
	strlen("\"") + 1)) == NULL) {
	return (rlc_return_error_ (inst->rlc_pi, RMDP_ALLOC_ERROR, NULL));
    }
    strcpy (tmp, "wm title . \"rmdp: ");
    strcat (tmp, name);
    strcat (tmp, "\"");
    rlc_gui_loadtcl(inst->rlc_pi, tmp);
    return(0);
#else
    return (rlc_return_error_ (inst->rlc_pi, RMDP_NO_UI, NULL));
#endif
}

/*++++
 Set the function to be called when the reception/transmission
 terminates successfully.

Params:
    - pi: protocol instance identifier.
    - end_f: the end callback function. It is of type
      typedef void (*rmdp_end_f) (int pi, int result);
 
Returns:
    - 0 on success;
    - a negative value on failure:
 */
int
rmdp_set_end_callback (int pi, rmdp_end_f end_f)
{
  rmdp_inst_t *inst;

    if ((inst = getinstance(pi)) == NULL) {
	return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }
    inst->end_f = end_f;

    return (0);
}

/*++++
 Enable receivers' stat reports.
 In the sender provides rmdp with a file descriptor to write
 receivers' stat reports. In the receiver enables stat reports
 generation if stafd > 0. If stafd < 0 generation/accumulation
 is disabled.

Params:
    - pi: protocol instance identifier.
    - stafd: In the sender: file descriptor. Must be writable.
	In the receiver enable (>0) and disable (<0) stats reports
	generation.
 
Returns:
    - 0 on success.
    - a negative value on failure:
*/
int
rmdp_set_stat_reports (int pi, int stafd)
{
  rmdp_inst_t *inst;
 
    if ((inst = getinstance(pi)) == NULL) {
        return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }

    inst->stafd = stafd;

    if (inst->isreceiver && inst->rlc_pi >= 0) {
	if (stafd > 0)
	    return rlc_rx_set_stats_deliver_f (inst->rlc_pi,
		pi, _rmdp_send_stats, RMDP_STATS_PAYSIZE);
	else
	    return rlc_rx_set_stats_deliver_f (inst->rlc_pi,
		pi, NULL, RMDP_STATS_PAYSIZE);
    }

    return ( 0 );
}

void _rmdp_deallocation(rmdp_inst_t *inst)
{
  int i, j;
    if (!inst) return;
    if (inst->alldata) {
	for (i=0; i<(inst->B); i++) {
	    if (inst->alldata[i])
		for (j=0; j<(inst->k); j++)
		    if (inst->alldata[i][j])
			free(inst->alldata[i][j]);
	    free(inst->alldata[i]);
	}
	free(inst->alldata);
    }
    if (inst->allpkts) {
	for (i=0; i<(inst->B); i++) {
	    if (inst->allpkts[i])
		free(inst->allpkts[i]);
	}
	free(inst->allpkts);
    }
    if (inst->whichdata) {
	for (i=0; i<(inst->B); i++) {
	    if (inst->whichdata[i])
		free(inst->whichdata[i]);
	}
	free(inst->whichdata);
    }
    if (inst->sparepkt)
	free(inst->sparepkt);
    if (inst->countdata)
	free(inst->countdata);
}

void _rmdp_deleteinstance(rmdp_inst_t *inst)
{
    if (!inst) return;

    if (inst->mfeeso != -1 && !inst->isreceiver) {
	rlc_deletechan(inst->mfeeso);
	close (inst->mfeeso);
	inst->mfeeso = -1;
    }
    if (inst->ufeeso != -1 && !inst->isreceiver) {
	rlc_deletechan(inst->ufeeso);
	close (inst->ufeeso);
	inst->ufeeso = -1;
    }

    _rmdp_deallocation(inst);

    _rlc_return_instance ((instance_des_t *) inst, rmdp_inst_mngr);
    free(inst);
}

int _rmdp_allocation(rmdp_inst_t *se)
{
  int i,j;
    /* allocate file buffers pointers only */
    if (NULL == (se->alldata = (void ***)calloc(se->B, sizeof(void *))) ||
	NULL == (se->allpkts = (void ***)calloc(se->B, sizeof(void *))) ||
	NULL == (se->countdata = (int *)calloc(se->B, sizeof(int))) ||
	NULL == (se->whichdata = (int **)calloc(se->B, sizeof(int *)))
	) {
	return (rlc_return_error_(se->rlc_pi, RMDP_ALLOC_ERROR,
	    "* rmdp: Cannot allocate file data structure"));
    }
    for (i=0; i<(se->B); i++) {
	if (NULL == (se->alldata[i] = (void **)calloc((se->k),
	    sizeof(void *))) ||
	    NULL == (se->allpkts[i] = (void **)calloc((se->k),
		sizeof(void *))) ||
	    NULL == (se->whichdata[i] = (int *)calloc((se->k),
		sizeof(int)))
	    ) {
	    return (rlc_return_error_(se->rlc_pi, RMDP_ALLOC_ERROR,
		"* rmdp: Cannot allocate file data structure"));
	}
	for (j=0; j<(se->k); j++) {
	    se->whichdata[i][j] = -1;
	    se->alldata[i][j] = se->allpkts[i][j] = NULL;
#if 0 /* NO HERE */
	    if (NULL ==(se->alldata[i][j]  =
		(char *)calloc(1, se->psize+RMDP_H_S))) {
		return (rlc_return_error_(se->rlc_pi, RMDP_ALLOC_ERROR,
		    "* rmdp: Cannot allocate file data structure"));
	    }
	    se->allpkts[i][j] = RMDP_DATA(se->alldata[i][j]);
	    if (!done && (read(fd, se->allpkts[i][j], se->psize)) !=
		se->psize) {
		done = 1;
	    }
#endif
	}
    }
    if (NULL ==(se->sparepkt = (char *)calloc(1, se->psize+RMDP_H_S))) {
	return (rlc_return_error_(se->rlc_pi, RMDP_ALLOC_ERROR,
	    "* rmdp: Cannot file allocate data structure"));
    }

    return (0);
}

